@@ -1,5 +1,5 @@
 /* pr -- convert text files for printing.
-   Copyright (C) 88, 91, 1995-1999 Free Software Foundation, Inc.
+   Copyright (C) 88, 91, 1995-2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -318,6 +318,7 @@
 #include <sys/types.h>
 #include <time.h>
 #include "system.h"
+#include "closeout.h"
 #include "error.h"
 #include "xstrtol.h"
 
@@ -725,11 +726,19 @@ static int last_line = FALSE;
    -h HEADER using pr test-suite */
 static int test_suite;
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  PAGES_OPTION = CHAR_MAX + 1,
+  COLUMNS_OPTION
+};
+
 static struct option const long_options[] =
 {
   {"test", no_argument, &test_suite, 1},
-  {"pages", required_argument, NULL, CHAR_MAX + 1},
-  {"columns", required_argument, NULL, CHAR_MAX + 2},
+  {"pages", required_argument, NULL, PAGES_OPTION},
+  {"columns", required_argument, NULL, COLUMNS_OPTION},
   {"across", no_argument, NULL, 'a'},
   {"show-control-chars", no_argument, NULL, 'c'},
   {"double-space", no_argument, NULL, 'd'},
@@ -845,63 +854,48 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  atexit (close_stdout);
+
   n_files = 0;
   file_names = (argc > 1
 		? (char **) xmalloc ((argc - 1) * sizeof (char *))
 		: NULL);
 
-  while (1)
+  while ((c = getopt_long (argc, argv,
+			   "-0123456789abcde::fFh:i::Jl:mn::N:o:rs::S::tTvw:W:",
+			   long_options, NULL))
+	 != -1)
     {
-      c = getopt_long (argc, argv,
-		       "-0123456789abcde::fFh:i::Jl:mn::N:o:rs::S::tTvw:W:",
-		       long_options, NULL);
-      if (c == 1)		/* Non-option argument. */
+      if (ISDIGIT (c))
 	{
-	  char *s;
-	  s = optarg;
-	  if (*s == '+')
-	    {
-	      /* long option --page dominates old `+FIRST_PAGE ...' */
-	      if (first_page_number > 0  || last_page_number > 0)
-		continue;
-	      ++s;
-	      first_last_page (s);
-	    }
-	  else
-	    {
-	      file_names[n_files++] = optarg;
-	    }
-	}
-      else
-	{
-	  if (ISDIGIT (c))
-	    {
-	      accum = accum * 10 + c - '0';
-	      continue;
-	    }
-	  else
-	    {
-	      if (accum > 0)
-		{
-		  columns = accum;
-		  explicit_columns = TRUE;
-		  accum = 0;
-		}
-	    }
+	  accum = accum * 10 + c - '0';
+	  continue;
 	}
 
-      if (c == 1)
-	continue;
-
-      if (c == EOF)
-	break;
+      if (accum > 0)
+	{
+	  columns = accum;
+	  explicit_columns = TRUE;
+	  accum = 0;
+	}
 
       switch (c)
 	{
 	case 0:			/* getopt long option */
 	  break;
 
-	case CHAR_MAX + 1:	/* --pages=FIRST_PAGE[:LAST_PAGE] */
+	case 1:			/* Non-option argument. */
+	  if (*optarg == '+')
+	    {
+	      /* long option --page dominates old `+FIRST_PAGE ...' */
+	      if (first_page_number <= 0 && last_page_number <= 0)
+		first_last_page (optarg);
+	    }
+	  else
+	    file_names[n_files++] = optarg;
+	  break;
+
+	case PAGES_OPTION:	/* --pages=FIRST_PAGE[:LAST_PAGE] */
 	  {			/* dominates old opt +... */
 	    if (optarg)
 	      first_last_page (optarg);
@@ -911,7 +905,7 @@ main (int argc, char **argv)
 	    break;
 	  }
 
-	case CHAR_MAX + 2:	/* --columns=COLUMN */
+	case COLUMNS_OPTION:	/* --columns=COLUMN */
 	  {
 	    long int tmp_long;
 	    if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
@@ -1154,8 +1148,6 @@ main (int argc, char **argv)
 
   if (have_read_stdin && fclose (stdin) == EOF)
     error (EXIT_FAILURE, errno, _("standard input"));
-  if (ferror (stdout) || fclose (stdout) == EOF)
-    error (EXIT_FAILURE, errno, _("write error"));
   if (failed_opens > 0)
     exit (EXIT_FAILURE);
   exit (EXIT_SUCCESS);
@@ -1636,6 +1628,15 @@ print_files (int number_of_files, char **av)
    time:  "yy-mm-dd HH:MM"  and:  "Page NNNN". */
 #define CHARS_FOR_DATE_AND_PAGE	25
 
+#define T_BUF_FMT "%Y-%m-%d %H:%M"	/* date/time short format */
+
+/* Add `2' because the expansion of %Y occupies 4 bytes, which is two more
+   than the length of `%Y'.  Each of the other formats expand to two bytes.  */
+#define T_BUF_SIZE (2 + sizeof T_BUF_FMT)
+
+/* This string is exactly the same length as the expansion of T_BUF_FMT.  */
+#define NO_DATE "-- Date/Time -- "
+
 /* Initialize header information.
    If DESC is non-negative, it is a file descriptor open to
    FILENAME for reading.
@@ -1648,28 +1649,28 @@ print_files (int number_of_files, char **av)
 static void
 init_header (char *filename, int desc)
 {
-  int chars_per_middle, chars_free, lhs_blanks, rhs_blanks;
   char *f = filename;
-  char *no_middle = "";
-  char *header_text;
-  struct tm *tmptr;
   struct stat st;
-  char *datim = "-- Date/Time -- ";
+  size_t header_buf_size;
 
   if (filename == NULL)
     f = "";
 
   if (header != NULL)
     free (header);
-  header = (char *) xmalloc (chars_per_line + 1);
+
+  /* Allow a space on each side of the the filename-or-header.  */
+  header_buf_size = MAX (chars_per_line, CHARS_FOR_DATE_AND_PAGE + 2) + 1;
+  header = (char *) xmalloc (header_buf_size);
 
   if (!standard_header && *custom_header == '\0')
     *header = '\0';			/* blank line header */
   else
     {
-#define T_BUF_FMT "%Y-%m-%d %H:%M"	/* date/time short format */
-#define T_BUF_SIZE 17  /* FIXME: using a literal here is fragile.  */
+      int chars_per_middle, lhs_blanks, rhs_blanks;
+      struct tm *tmptr;
       char t_buf[T_BUF_SIZE];
+      char *header_text;
 
       /* If parallel files or standard input, use current time. */
       if (desc < 0 || STREQ (filename, "-") || fstat (desc, &st))
@@ -1678,15 +1679,16 @@ init_header (char *filename, int desc)
       tmptr = localtime (&st.st_mtime);
       strftime (t_buf, T_BUF_SIZE, T_BUF_FMT, tmptr);
 
-      chars_per_middle = chars_per_line - CHARS_FOR_DATE_AND_PAGE;
+      chars_per_middle = header_buf_size - 1 - CHARS_FOR_DATE_AND_PAGE;
       if (chars_per_middle < 3)
 	{
-	  header_text = no_middle;	/* Nothing free for a heading */
+	  header_text = "";	/* Nothing free for a heading */
 	  lhs_blanks = 1;
 	  rhs_blanks = 1;
 	}
       else
 	{
+	  int chars_free;
 	  header_text = standard_header ? f : custom_header;
 	  chars_free = chars_per_middle - (int) strlen (header_text);
 	  if (chars_free > 1)
@@ -1703,7 +1705,7 @@ init_header (char *filename, int desc)
 	    }
 	}
 
-      sprintf (header, _("%s%*s%s%*sPage"), (test_suite ? datim : t_buf),
+      sprintf (header, _("%s%*s%s%*sPage"), (test_suite ? NO_DATE : t_buf),
 	       lhs_blanks, " ", header_text, rhs_blanks, " ");
     }
 }
@@ -2009,7 +2011,7 @@ store_columns (void)
   /* Keep track of the location of the last char in buff. */
   line_vector[line] = buff_start;
 
-  if (balance_columns && p->lines_stored != lines_per_body)
+  if (balance_columns)
     balance (line);
 }
 
@@ -2445,16 +2447,12 @@ print_header (void)
 static int
 read_line (COLUMN *p)
 {
-  register int c, chars;
+  int c;
+  int chars IF_LINT (= 0);
   int last_input_position;
   int j, k;
   COLUMN *q;
 
-/* Suppress `used before initialized' warning. */
-#ifdef lint
-  chars = 0;
-#endif
-
   /* read 1st character in each line or any character succeeding a FF: */
   c = getc (p->fp);
 
