@@ -171,6 +171,10 @@ char *alloca ();
 # include <stdint.h>
 #endif
 
+#if HAVE_INTTYPES_H
+# include <inttypes.h> /* for the definition of UINTMAX_MAX */
+#endif
+
 #include <ctype.h>
 
 /* Jim Meyering writes:
@@ -527,6 +531,10 @@ enum
 # define SIZE_MAX TYPE_MAXIMUM (size_t)
 #endif
 
+#ifndef UINTMAX_MAX
+# define UINTMAX_MAX TYPE_MAXIMUM (uintmax_t)
+#endif
+
 #ifndef OFF_T_MIN
 # define OFF_T_MIN TYPE_MINIMUM (off_t)
 #endif
